from eventsourcing.application.simple import ApplicationWithConcreteInfrastructure
from eventsourcing.infrastructure.axonserver.factory import AxonInfrastructureFactory


class AxonApplication(ApplicationWithConcreteInfrastructure):
    infrastructure_factory_class = AxonInfrastructureFactory
